// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_continuous_deployment/src/core/commands/build_app/build_app_task.dart';
import 'package:wyatt_continuous_deployment/src/core/constants/app_constants.dart';
import 'package:wyatt_continuous_deployment/src/core/injection/injectable.dart';
import 'package:wyatt_continuous_deployment/src/core/task/task_cubit.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/context.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/decrypt_keys_parameters.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/fastlane.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/fastlane/fastlane_create_keychain_parameters.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/fastlane/fastlane_get_certificates_parameters.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/config/get_ios_config_usecase.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/decrypt_keys/decrypt_keys_usecase.dart';

part 'build_ipa_state.dart';

class BuildIpaTask extends BuildAppTask {
  factory BuildIpaTask() => BuildIpaTask._(
        flutter: getIt(),
        fastlane: getIt(),
        decryptKeysUsecase: getIt(),
        getIosConfigUsecase: getIt(),
        checkToolsUsecase: getIt(),
        checkConfigFileUsecase: getIt(),
      );

  BuildIpaTask._({
    required DecryptKeysUsecase decryptKeysUsecase,
    required GetIosConfigUsecase getIosConfigUsecase,
    required Fastlane fastlane,
    required super.flutter,
    required super.checkToolsUsecase,
    required super.checkConfigFileUsecase,
  })  : _decryptKeysUsecase = decryptKeysUsecase,
        _getIosConfigUsecase = getIosConfigUsecase,
        _fastlane = fastlane;

  final DecryptKeysUsecase _decryptKeysUsecase;
  final GetIosConfigUsecase _getIosConfigUsecase;
  final Fastlane _fastlane;

  @override
  FutureOr<void> buildApp(Context ctx) async {
    // 1. Get iOS Config
    emit(BuildIpaStateGetIosConfig());
    final getIosConfigResult = await _getIosConfigUsecase(ctx.configFile.path);
    if (getIosConfigResult.isErr) {
      return emit(
        TaskStateError(message: getIosConfigResult.err.toString()),
      );
    }

    // 2. Decrypt iOS keys
    emit(BuildIpaStateDecryptIosKey());
    final decryptIosKeyResult = await _decryptKeysUsecase(
      DecryptKeysParameters(
        encryptedFilePath: getIosConfigResult.ok!.encryptedKeyFile.path,
        passphrase:
            ctx.environment[AppConstants.iosKeysSecretPassphrase].toString(),
      ),
    );
    if (decryptIosKeyResult.isErr) {
      return emit(
        TaskStateError(message: decryptIosKeyResult.err.toString()),
      );
    }

    // 5. Create Keychain
    emit(BuildIpaStateCreateKeychain());
    final fastlaneCreateKeychainResult = await _fastlane.createKeychain(
      FastlaneCreateKeychainParameters(
        name: 'match',
        password: 'flutter-cd-cli',
        gitPassphrase: getIosConfigResult.ok!.gitPassphrase,
      ),
    );
    if (fastlaneCreateKeychainResult.isErr) {
      return emit(
        TaskStateError(message: fastlaneCreateKeychainResult.err.toString()),
      );
    }

    // 6. Retrieve certificates
    emit(BuildIpaStateRetrieveCertificates());
    final fastlaneGetCertificateResult = await _fastlane.getCertificates(
      FastlaneGetCertificatesParameters(
        appIdentifier: getIosConfigResult.ok!.developerAppIdentifier,
        gitPrivateKey:
            '${getIosConfigResult.ok!.iosFolderPath}/${getIosConfigResult.ok!.gitCertsPrivateKeyFileName}',
        keychainName: 'match',
        keychainPassword: 'flutter-cd-cli',
        gitUrl: getIosConfigResult.ok!.gitUrl,
        gitPassphrase: getIosConfigResult.ok!.gitPassphrase,
      ),
    );
    if (fastlaneGetCertificateResult.isErr) {
      return emit(
        TaskStateError(message: fastlaneGetCertificateResult.err.toString()),
      );
    }

    // 7. Build IPA
    emit(BuildIpaStateBuildIpa());
    final flutterBuildIpaResult =
        await flutter.buildIpa(getIosConfigResult.ok!.flutterArgs.split(' '));
    if (flutterBuildIpaResult.isErr) {
      return emit(
        TaskStateError(message: flutterBuildIpaResult.err.toString()),
      );
    }

    emit(const TaskStateSucceed());
  }
}
