// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:wyatt_crud_bloc/wyatt_crud_bloc.dart' hide Query;
import 'package:wyatt_crud_bloc/wyatt_crud_bloc.dart' as crud show Query;

/// {@template crud_data_source_firestore_impl}
/// A concrete implementation of [CrudDataSource] that uses
/// [FirebaseFirestore] as the data source.
/// {@endtemplate}
class CrudDataSourceFirestoreImpl extends CrudDataSource {
  /// {@macro crud_data_source_firestore_impl}
  CrudDataSourceFirestoreImpl(
    String collection, {
    required super.id,

    /// The function that converts a [DocumentSnapshot] to a Model.
    required Map<String, dynamic> Function(
      DocumentSnapshot<Map<String, dynamic>>,
      SnapshotOptions?,
    ) fromFirestore,

    /// The function that converts a Model to a [Map<String, Object?>].
    required Map<String, Object?> Function(Map<String, dynamic>, SetOptions?)
        toFirestore,
    FirebaseFirestore? firestore,
  }) : _firestore = firestore ?? FirebaseFirestore.instance {
    _collectionReference =
        _firestore.collection(collection).withConverter<Map<String, dynamic>>(
              fromFirestore: fromFirestore,
              toFirestore: toFirestore,
            );
  }
  final FirebaseFirestore _firestore;
  late CollectionReference<Map<String, dynamic>> _collectionReference;

  @override
  Future<void> create(Map<String, dynamic> object, {String? id}) {
    if (id != null) {
      return _collectionReference.doc(id).set(object);
    } else {
      final String? id = this.id(
        OperationType.create,
        object,
      );

      if (id != null) {
        return _collectionReference.doc(id).set(object);
      }
      return _collectionReference.add(object);
    }
  }

  @override
  Future<void> delete(String id) => _collectionReference.doc(id).delete();

  @override
  Future<void> deleteAll() async {
    final batch = _firestore.batch();
    final QuerySnapshot snapshots = await _collectionReference.get();
    for (final DocumentSnapshot snapshot in snapshots.docs) {
      batch.delete(snapshot.reference);
    }
    return batch.commit();
  }

  @override
  Future<Map<String, dynamic>?> get(String id) async {
    final DocumentSnapshot<Map<String, dynamic>> snapshot =
        await _collectionReference.doc(id).get();
    return snapshot.data();
  }

  @override
  Future<List<Map<String, dynamic>?>> getAll() async {
    final QuerySnapshot<Map<String, dynamic>> snapshots =
        await _collectionReference.get();
    return snapshots.docs.map((snapshot) => snapshot.data()).toList();
  }

  @override
  Future<List<Map<String, dynamic>?>> search(
    List<crud.Query> conditions,
  ) async {
    Query<Map<String, dynamic>> query = _collectionReference;
    for (final condition in conditions) {
      query = _queryParser(condition, query);
    }
    final QuerySnapshot<Map<String, dynamic>> snapshots = await query.get();
    return snapshots.docs.map((snapshot) => snapshot.data()).toList();
  }

  @override
  Stream<List<Map<String, dynamic>?>> stream({
    String? id,
    List<crud.Query>? conditions,
    bool includeMetadataChanges = false,
  }) {
    if (id != null) {
      return _collectionReference
          .doc(id)
          .snapshots(
            includeMetadataChanges: includeMetadataChanges,
          )
          .map<List<Map<String, dynamic>?>>(
            (snapshot) => [snapshot.data()],
          );
    } else {
      if (conditions != null) {
        Query<Map<String, dynamic>> query = _collectionReference;
        for (final condition in conditions) {
          query = _queryParser(condition, query);
        }
        return query
            .snapshots(
              includeMetadataChanges: includeMetadataChanges,
            )
            .map(
              (querySnapshot) => querySnapshot.docs
                  .map((snapshot) => snapshot.data())
                  .toList(),
            );
      } else {
        return _collectionReference
            .snapshots(
              includeMetadataChanges: includeMetadataChanges,
            )
            .map(
              (querySnapshot) => querySnapshot.docs
                  .map((snapshot) => snapshot.data())
                  .toList(),
            );
      }
    }
  }

  @override
  Future<void> update(
    String id, {
    Map<String, dynamic>? object,
  }) {
    if (object != null) {
      return _collectionReference.doc(id).update(object);
    } else {
      throw Exception('You must provide data to update');
    }
  }

  @override
  Future<void> updateAll(Map<String, Object?>? data) async {
    if (data == null) {
      throw Exception('You must provide data to update');
    }
    final batch = _firestore.batch();
    final QuerySnapshot<Map<String, dynamic>> snapshots =
        await _collectionReference.get();
    for (final DocumentSnapshot snapshot in snapshots.docs) {
      batch.update(snapshot.reference, data);
    }
    return batch.commit();
  }

  Query<Map<String, dynamic>> _queryParser(crud.Query condition, Object query) {
    query as Query<Map<String, dynamic>>;
    if (condition is WhereQuery) {
      switch (condition.type) {
        case WhereQueryType.isEqualTo:
          return query.where(condition.field, isEqualTo: condition.value);
        case WhereQueryType.isNotEqualTo:
          return query.where(condition.field, isNotEqualTo: condition.value);
        case WhereQueryType.isLessThan:
          return query.where(condition.field, isLessThan: condition.value);
        case WhereQueryType.isLessThanOrEqualTo:
          return query.where(
            condition.field,
            isLessThanOrEqualTo: condition.value,
          );
        case WhereQueryType.isGreaterThan:
          return query.where(condition.field, isGreaterThan: condition.value);
        case WhereQueryType.isGreaterThanOrEqualTo:
          return query.where(
            condition.field,
            isGreaterThanOrEqualTo: condition.value,
          );
        case WhereQueryType.arrayContains:
          return query.where(condition.field, arrayContains: condition.value);
        case WhereQueryType.arrayContainsAny:
          return query.where(
            condition.field,
            arrayContainsAny: condition.value as List<Object>,
          );
        case WhereQueryType.whereIn:
          return query.where(
            condition.field,
            whereIn: condition.value as List<Object>,
          );
        case WhereQueryType.whereNotIn:
          return query.where(
            condition.field,
            whereNotIn: condition.value as List<Object>,
          );
        case WhereQueryType.isNull:
          return query.where(condition.field, isNull: condition.value as bool);
      }
    } else if (condition is LimitQuery) {
      return query.limit(condition.limit);
    } else if (condition is OrderByQuery) {
      return query.orderBy(
        condition.field,
        descending: !condition.ascending,
      );
    }
    return query;
  }
}
